@@ -133,6 +133,18 @@ int create_comm_struct (struct perftest_comm *comm,
 		struct perftest_parameters *user_param);
 
 
+/* dealloc_comm_struct .
+ *
+ * Description : deallocating the already allocated structs in perftest_comm.
+ *
+ * Parameters :
+ *  comm - A Communication struct.
+ *  user_param -  Perftest parameters.
+ */
+void dealloc_comm_struct(struct perftest_comm *comm,
+		struct perftest_parameters *user_param);
+
+
 /* set_up_connection .
  *
  * Description : Fills the my_dest with all of the machine proporties.
@@ -409,6 +421,14 @@ void xchg_bw_reports (struct perftest_comm *comm, struct bw_report_data *my_bw_r
  */
 void exchange_versions (struct perftest_comm *user_comm, struct perftest_parameters *user_param);
 
+/* check_version_compatibility.
+ *
+ * Description :
+ * 	Check compatibility between old and new implementation.
+ *
+ */
+void check_version_compatibility(struct perftest_parameters *user_param);
+
 /* check_sys_data.
  *
  * Description :
