@@ -74,6 +74,7 @@ static int set_mcast_group(struct pingpong_context *ctx,
 	mcg_params->sm_sl   = port_attr.sm_sl;
 	mcg_params->ib_port = user_param->ib_port;
 	mcg_params->ib_ctx  = ctx->context;
+	mcg_params->ib_devname = user_param->ib_devname;
 
 	if (!strcmp(link_layer_str(user_param->link_type),"IB")) {
 		/* Request for Mcast group create registery in SM. */
@@ -94,7 +95,6 @@ static int send_set_up_connection(struct pingpong_context *ctx,
 		struct mcast_parameters *mcg_params,
 		struct perftest_comm *comm)
 {
-	int i;
 
 	if (set_up_connection(ctx,user_param,my_dest)) {
 		fprintf(stderr," Unable to set up my IB connection parameters\n");
@@ -102,6 +102,7 @@ static int send_set_up_connection(struct pingpong_context *ctx,
 	}
 
 	if (user_param->use_mcg) {
+		int i;
 		mcg_params->user_mgid = user_param->user_mgid;
 
 		set_multicast_gid(mcg_params,ctx->qp[0]->qp_num,(int)user_param->machine);
@@ -131,8 +132,8 @@ static int send_destroy_ctx(struct pingpong_context *ctx,
 		struct perftest_parameters *user_param,
 		struct mcast_parameters *mcg_params)
 {
-	int i;
 	if (user_param->use_mcg) {
+		int i;
 		for (i=0; i < user_param->num_of_qps; i++) {
 				if (ibv_detach_mcast(ctx->qp[i],&mcg_params->base_mgid,mcg_params->base_mlid)) {
 					fprintf(stderr, "Couldn't dettach QP to MultiCast group\n");
@@ -162,7 +163,7 @@ static int send_destroy_ctx(struct pingpong_context *ctx,
  ******************************************************************************/
 int main(int argc, char *argv[])
 {
-	int                        i = 0, rc;
+	int                        i = 0, rc, error = 1;
 	int                        size_max_pow = 24;
 	int			   ret_val;
 	struct report_options      report;
@@ -211,7 +212,7 @@ int main(int argc, char *argv[])
 	}
 
 	/* Getting the relevant context from the device */
-	ctx.context = ibv_open_device(ib_dev);
+	ctx.context = ctx_open_device(ib_dev, &user_param);
 	if (!ctx.context) {
 		fprintf(stderr, " Couldn't get context for the device\n");
 		return FAILURE;
@@ -244,26 +245,32 @@ int main(int argc, char *argv[])
 	/* Initialize the connection and print the local data. */
 	if (establish_connection(&user_comm)) {
 		fprintf(stderr," Unable to init the socket connection\n");
+		dealloc_comm_struct(&user_comm,&user_param);
 		return FAILURE;
 	}
 
 	exchange_versions(&user_comm, &user_param);
-
+	check_version_compatibility(&user_param);
 	check_sys_data(&user_comm, &user_param);
 
 	/* See if MTU and link type are valid and supported. */
 	if (check_mtu(ctx.context,&user_param, &user_comm)) {
 		fprintf(stderr, " Couldn't get context for the device\n");
+		dealloc_comm_struct(&user_comm,&user_param);
 		return FAILURE;
 	}
 
-	ALLOCATE(my_dest , struct pingpong_dest , user_param.num_of_qps);
+	MAIN_ALLOC(my_dest , struct pingpong_dest , user_param.num_of_qps , return_error);
 	memset(my_dest, 0, sizeof(struct pingpong_dest)*user_param.num_of_qps);
-	ALLOCATE(rem_dest , struct pingpong_dest , user_param.num_of_qps);
+	MAIN_ALLOC(rem_dest , struct pingpong_dest , user_param.num_of_qps , free_my_dest);
 	memset(rem_dest, 0, sizeof(struct pingpong_dest)*user_param.num_of_qps);
 
 	/* Allocating arrays needed for the test. */
-	alloc_ctx(&ctx,&user_param);
+	if(alloc_ctx(&ctx,&user_param)){
+		fprintf(stderr, "Couldn't allocate context\n");
+		dealloc_comm_struct(&user_comm,&user_param);
+		goto free_mem;
+	}
 
 	/* Create RDMA CM resources and connect through CM. */
 	if (user_param.work_rdma_cm == ON) {
@@ -272,45 +279,43 @@ int main(int argc, char *argv[])
 		if (rc) {
 			fprintf(stderr,
 				"Failed to create RDMA CM connection with resources.\n");
-			return FAILURE;
+			dealloc_comm_struct(&user_comm,&user_param);
+			dealloc_ctx(&ctx, &user_param);
+			goto free_mem;
 		}
 	} else {
 		/* create all the basic IB resources (data buffer, PD, MR, CQ and events channel) */
 		if (ctx_init(&ctx, &user_param)) {
 			fprintf(stderr, " Couldn't create IB resources\n");
-			return FAILURE;
+			dealloc_comm_struct(&user_comm,&user_param);
+			dealloc_ctx(&ctx, &user_param);
+			goto free_mem;
 		}
 	}
 
 	/* Set up the Connection. */
 	if (send_set_up_connection(&ctx,&user_param,my_dest,&mcg_params,&user_comm)) {
 		fprintf(stderr," Unable to set up socket connection\n");
-		return FAILURE;
+		goto destroy_ctx;
 	}
 
 	/* Print basic test information. */
 	ctx_print_test_info(&user_param);
 
-	for (i=0; i < user_param.num_of_qps; i++)
-		ctx_print_pingpong_data(&my_dest[i],&user_comm);
-
-	user_comm.rdma_params->side = REMOTE;
 	for (i=0; i < user_param.num_of_qps; i++) {
 
 		/* shaking hands and gather the other side info. */
 		if (ctx_hand_shake(&user_comm,&my_dest[i],&rem_dest[i])) {
 			fprintf(stderr,"Failed to exchange data between server and clients\n");
-			return FAILURE;
+			goto destroy_ctx;
 		}
-
-		ctx_print_pingpong_data(&rem_dest[i],&user_comm);
 	}
 
 	if (user_param.work_rdma_cm == OFF) {
 		if (ctx_check_gid_compatibility(&my_dest[0], &rem_dest[0])) {
 			fprintf(stderr,"\n Found Incompatibility issue with GID types.\n");
 			fprintf(stderr," Please Try to use a different IP version.\n\n");
-			return FAILURE;
+			goto destroy_ctx;
 		}
 	}
 
@@ -323,7 +328,7 @@ int main(int argc, char *argv[])
 			/* Request for Mcast group create registery in SM. */
 			if (join_multicast_group(SUBN_ADM_METHOD_SET,&mcg_params)) {
 				fprintf(stderr," Failed to Join Mcast request\n");
-				return FAILURE;
+				goto destroy_ctx;
 			}
 		}
 
@@ -344,26 +349,52 @@ int main(int argc, char *argv[])
 		/* Prepare IB resources for rtr/rts. */
 		if (ctx_connect(&ctx,rem_dest,&user_param,my_dest)) {
 			fprintf(stderr," Unable to Connect the HCA's through the link\n");
-			return FAILURE;
+			goto destroy_ctx;
 		}
 	}
 
 	/* shaking hands and gather the other side info. */
 	if (ctx_hand_shake(&user_comm,&my_dest[0],&rem_dest[0])) {
 		fprintf(stderr,"Failed to exchange data between server and clients\n");
-		return FAILURE;
+		goto destroy_ctx;
+	}
+
+	if (user_param.connection_type == DC)
+	{
+		/* Set up connection one more time to send qpn properly for DC */
+		if (set_up_connection(&ctx, &user_param, my_dest))
+		{
+			fprintf(stderr, " Unable to set up socket connection\n");
+			goto destroy_ctx;
+		}
+	}
+
+	/* Print this machine QP information */
+	for (i=0; i < user_param.num_of_qps; i++)
+		ctx_print_pingpong_data(&my_dest[i],&user_comm);
+
+	user_comm.rdma_params->side = REMOTE;
+
+	for (i=0; i < user_param.num_of_qps; i++) {
+
+		if (ctx_hand_shake(&user_comm,&my_dest[i],&rem_dest[i])) {
+			fprintf(stderr," Failed to exchange data between server and clients\n");
+			goto destroy_ctx;
+		}
+
+		ctx_print_pingpong_data(&rem_dest[i],&user_comm);
 	}
 
 	if (user_param.use_event) {
 
 		if (ibv_req_notify_cq(ctx.send_cq, 0)) {
 			fprintf(stderr, "Couldn't request RCQ notification\n");
-			return FAILURE;
+			goto destroy_ctx;
 		}
 
 		if (ibv_req_notify_cq(ctx.recv_cq, 0)) {
 			fprintf(stderr, "Couldn't request RCQ notification\n");
-			return FAILURE;
+			goto destroy_ctx;
 		}
 	}
 	if (user_param.output == FULL_VERBOSITY) {
@@ -387,7 +418,7 @@ int main(int argc, char *argv[])
 			/* Post receive recv_wqes fo current message size */
 			if (ctx_set_recv_wqes(&ctx,&user_param)) {
 				fprintf(stderr," Failed to post receive recv_wqes\n");
-				return FAILURE;
+				goto free_mem;
 			}
 
 			/* Sync between the client and server so the client won't send packets
@@ -396,11 +427,13 @@ int main(int argc, char *argv[])
 
 			if (ctx_hand_shake(&user_comm,&my_dest[0],&rem_dest[0])) {
 				fprintf(stderr,"Failed to exchange data between server and clients\n");
-				return FAILURE;
+				goto free_mem;
 			}
 
-			if(run_iter_lat_send(&ctx, &user_param))
-				return 17;
+			if(run_iter_lat_send(&ctx, &user_param)){
+				error = 17;
+				goto free_mem;
+			}
 
 			user_param.test_type == ITERATIONS ? print_report_lat(&user_param) : print_report_lat_duration(&user_param);
 		}
@@ -410,7 +443,7 @@ int main(int argc, char *argv[])
 		/* Post recevie recv_wqes fo current message size */
 		if (ctx_set_recv_wqes(&ctx,&user_param)) {
 			fprintf(stderr," Failed to post receive recv_wqes\n");
-			return FAILURE;
+			goto free_mem;
 		}
 
 		/* Sync between the client and server so the client won't send packets
@@ -419,11 +452,13 @@ int main(int argc, char *argv[])
 
 		if (ctx_hand_shake(&user_comm,my_dest,rem_dest)) {
 			fprintf(stderr,"Failed to exchange data between server and clients\n");
-			return FAILURE;
+			goto free_mem;
 		}
 
-		if(run_iter_lat_send(&ctx, &user_param))
-			return 17;
+		if(run_iter_lat_send(&ctx, &user_param)){
+			error = 17;
+			goto free_mem;
+		}
 
 		user_param.test_type == ITERATIONS ? print_report_lat(&user_param) : print_report_lat_duration(&user_param);
 	}
@@ -437,5 +472,36 @@ int main(int argc, char *argv[])
 		fprintf(stderr," Trying to close this side resources\n");
 	}
 
+	if (user_param.work_rdma_cm == ON) {
+		if (send_destroy_ctx(&ctx,&user_param,&mcg_params)) {
+			fprintf(stderr, "Failed to destroy resources\n");
+			goto destroy_cm_context;
+		}
+		user_comm.rdma_params->work_rdma_cm = OFF;
+		free(rem_dest);
+		free(my_dest);
+		return destroy_ctx(user_comm.rdma_ctx,user_comm.rdma_params);
+	}
+
+	free(rem_dest);
+	free(my_dest);
+
 	return send_destroy_ctx(&ctx,&user_param,&mcg_params);
+
+destroy_ctx:
+	if (send_destroy_ctx(&ctx,&user_param,&mcg_params))
+		fprintf(stderr, "Failed to destroy resources\n");
+
+destroy_cm_context:
+	if (user_param.work_rdma_cm == ON) {
+		user_comm.rdma_params->work_rdma_cm = OFF;
+		destroy_ctx(user_comm.rdma_ctx,user_comm.rdma_params);
+	}
+
+free_mem:
+	free(rem_dest);
+free_my_dest:
+	free(my_dest);
+return_error:
+	return error;
 }
